#!/usr/bin/python3
"""
Plot CO2 levels vs. wind direction.

Input data is rather special to me; what might be generally useful is the
interpolation scheme for the wind direction and speed.

In case you want to re-use any of this: Distributed under CC-0 (i.e.
public domain for all intents and purposes).
"""

import contextlib
import csv
import math
import sys
import sqlite3 as sqlite

import numpy
from matplotlib import pyplot


@contextlib.contextmanager
def fig_writeto(dest_name, **kwargs):
	"""starts a new pyplot figure and saves it to dest_name once
	the controlled section is left.
	"""
	fig = pyplot.figure(figsize=(4,4))
	ax = fig.add_subplot(111, **kwargs)
	yield ax

	if dest_name is None:
		pyplot.show()
	else:
		fig.savefig(dest_name, dpi=200)


class Weather:
	"""An interpolating interface to a record from weathercheck.py

	See https://blog.tfiu.de/media/2021/weathercheck.py for where the
	DB comes from.

	And yes, this won't scale: one db query per point isn't any good for
	large plots.  But mine isn't large.
	"""
	def __init__(self, db_path):
		self.conn = sqlite.connect(db_path,
			detect_types=sqlite.PARSE_DECLTYPES|sqlite.PARSE_COLNAMES)

	def get_for_time(self, time, col_name, angle_fix=False, default=None):
		"""returns the value col_name linearly interpolated for time.

		This raises a ValueError if the column has a NULL for at least one of
		the surrounding dates (unless default is passed, in which case
		that is returned).

		Time must be a unit timestamp.

		If angle_fix is true, we will try to catch cases straddling the stitching
		line for angles in degrees and deal sensibly with it.

		WARNING: col_name allows SQL injection.  Don't allow external strings 
		in there.
		"""
		res = list(self.conn.execute(f"SELECT timestamp, {col_name} FROM climate"
			" WHERE TIMESTAMP BETWEEN ? AND ?"
			" ORDER BY ABS(timestamp-?) LIMIT 2",
			(time-40000, time+40000, time)))

		if len(res)!=2:
			if default is not None:
				return default
			raise ValueError(f"No data points close to {time}")

		elif abs(res[0][0]-time)<200 and res[0][1] is not None:
			#	direct hit
			return res[0][1]

		elif res[0][1] is None or res[1][1] is None:
			if default is not None:
				return default
			raise ValueError("One or more limits missing.  Cannot interpolate.")

		else:
			t1, v1 = res[0]
			t2, v2 = res[1]

			if angle_fix:
				if abs(v1-v2)>180:
					# interpolate over 0 rather than 180 degrees
					if v1>v2:
						v2 += 360
					else:
						v1 += 360
				return (v1*(t2-time)+v2*(time-t1))/(t2-t1)%360

			else:
				return (v1*(t2-time)+v2*(time-t1))/(t2-t1)


def get_conc_speed_dirs(co2_name, db_path, limit_to=1000000):
	"""returns concentrations and wind speeds and directions for those points in 
	co2_name that have wind_speeds in db_path.

	Use limit_to to restrict to some smallish number of records from co2_name.
	"""
	weather = Weather(db_path)

	times, angles, speeds, concs = [], [], [], []
	with open(co2_name) as in_file:
		for line in in_file:
			if not line.strip() or line.startswith("#"):
				pass
			time, co2, temp = line.split()
			time = int(float(time))
			try:
				angles.append(
					weather.get_for_time(time, "wind_dir", True)/180*math.pi)
				speeds.append(weather.get_for_time(time, "wind_speed", default=1))
				concs.append(float(co2))
				times.append(time)
			except ValueError:
				# no wind direction available, skip data point
				pass

			limit_to -= 1
			if limit_to==0:
				break

	return times, angles, speeds, concs

def main(co2_name, db_path):
	times, angles, speeds, concs = get_conc_speed_dirs(
		co2_name, db_path, limit_to=1000000)

	with open("winds.csv", "w") as f:
		writer = csv.writer(f)
		writer.writerow(["time", "angle", "speed", "conc"])
		writer.writerows(zip(times, angles, speeds, concs))

	with fig_writeto("angle-conc-speed.png", projection="polar") as ax:
		ax.set_theta_direction(-1)
		ax.set_theta_zero_location("N")
		ax.set_ylim(ymin=300, ymax=max(concs))
		pyplot.set_cmap("Oranges")
		sc = ax.scatter(angles, concs, c=speeds, s=1)
		pyplot.colorbar(sc)


if __name__=="__main__":
	if len(sys.argv)!=3:
		sys.exit(f"Usage: {sys.argv[0]} <co2-wsv> <weather-sqlite>")
	main(sys.argv[1], sys.argv[2])
